@@ -1,20 +1,33 @@
-from __future__ import division, print_function
-
+# -----------------------------------------------------------------------------
+# Copyright 2020-2021, Harald Lieder, mailto:harald.lieder@outlook.com
+# License: GNU AFFERO GPL 3.0, https://www.gnu.org/licenses/agpl-3.0.html
+# Part of "PyMuPDF", Python bindings for "MuPDF" (http://mupdf.com), a
+# lightweight PDF, XPS, and E-book viewer, renderer and toolkit which is
+# maintained and developed by Artifex Software, Inc. https://artifex.com.
+# -----------------------------------------------------------------------------
+import argparse
+import bisect
 import os
 import sys
+import statistics
+from typing import Dict, List, Set, Tuple
 
 import fitz
+from fitz.fitz import (
+    TEXT_INHIBIT_SPACES,
+    TEXT_PRESERVE_LIGATURES,
+    TEXT_PRESERVE_WHITESPACE,
+)
 
 mycenter = lambda x: (" %s " % x).center(75, "-")
 
 
 def recoverpix(doc, item):
-    """Return image for a given XREF.
-    """
+    """Return image for a given XREF."""
     x = item[0]  # xref of PDF image
     s = item[1]  # xref of its /SMask
     if s == 0:  # no smask: use direct image output
-        return doc.extractImage(x)
+        return doc.extract_image(x)
 
     def getimage(pix):
         if pix.colorspace.n != 4:
@@ -38,7 +51,7 @@ def recoverpix(doc, item):
         return getimage(pix1)  # return the pixmap as is
 
     pix = fitz.Pixmap(pix1)  # copy of pix1, with an alpha channel added
-    pix.setAlpha(pix2.samples)  # treat pix2.samples as the alpha values
+    pix.set_alpha(pix2.samples)  # treat pix2.samples as the alpha values
     pix1 = pix2 = None  # free temp pixmaps
 
     # we may need to adjust something for CMYK pixmaps here:
@@ -46,13 +59,12 @@ def recoverpix(doc, item):
 
 
 def open_file(filename, password, show=False, pdf=True):
-    """Open and authenticate a document.
-    """
+    """Open and authenticate a document."""
     doc = fitz.open(filename)
-    if not doc.isPDF and pdf is True:
+    if not doc.is_pdf and pdf is True:
         sys.exit("this command supports PDF files only")
     rc = -1
-    if not doc.needsPass:
+    if not doc.needs_pass:
         return doc
     if password:
         rc = doc.authenticate(password)
@@ -66,8 +78,7 @@ def open_file(filename, password, show=False, pdf=True):
 
 
 def print_dict(item):
-    """Print a Python dictionary.
-    """
+    """Print a Python dictionary."""
     l = max([len(k) for k in item.keys()]) + 1
     for k, v in item.items():
         msg = "%s: %s" % (k.rjust(l), v)
@@ -82,9 +93,9 @@ def print_xref(doc, xref):
     For a stream also print its size.
     """
     print("%i 0 obj" % xref)
-    xref_str = doc.xrefObject(xref)
+    xref_str = doc.xref_object(xref)
     print(xref_str)
-    if doc.isStream(xref):
+    if doc.xref_is_stream(xref):
         temp = xref_str.split()
         try:
             idx = temp.index("/Length") + 1
@@ -158,28 +169,28 @@ def show(args):
         "'%s', pages: %i, objects: %i, %g %s, %s, encryption: %s"
         % (
             args.input,
-            doc.pageCount,
-            doc._getXrefLength() - 1,
+            doc.page_count,
+            doc.xref_length() - 1,
             size,
             flag,
             meta["format"],
             meta["encryption"],
         )
     )
-    n = doc.isFormPDF
+    n = doc.is_form_pdf
     if n > 0:
-        s = doc.getSigFlags()
+        s = doc.get_sigflags()
         print(
             "document contains %i root form fields and is %ssigned"
             % (n, "not " if s != 3 else "")
         )
-    n = doc.embeddedFileCount()
+    n = doc.embfile_count()
     if n > 0:
         print("document contains %i embedded files" % n)
     print()
     if args.catalog:
         print(mycenter("PDF catalog"))
-        xref = doc.PDFCatalog()
+        xref = doc.pdf_catalog()
         print_xref(doc, xref)
         print()
     if args.metadata:
@@ -188,22 +199,22 @@ def show(args):
         print()
     if args.xrefs:
         print(mycenter("object information"))
-        xrefl = get_list(args.xrefs, doc._getXrefLength(), what="xref")
+        xrefl = get_list(args.xrefs, doc.xref_length(), what="xref")
         for xref in xrefl:
             print_xref(doc, xref)
             print()
     if args.pages:
         print(mycenter("page information"))
-        pagel = get_list(args.pages, doc.pageCount + 1)
+        pagel = get_list(args.pages, doc.page_count + 1)
         for pno in pagel:
             n = pno - 1
-            xref = doc._getPageXref(n)[0]
+            xref = doc.page_xref(n)
             print("Page %i:" % pno)
             print_xref(doc, xref)
             print()
     if args.trailer:
         print(mycenter("PDF trailer"))
-        print(doc.PDFTrailer())
+        print(doc.pdf_trailer())
         print()
     doc.close()
 
@@ -232,11 +243,11 @@ def clean(args):
         return
 
     # create sub document from page numbers
-    pages = get_list(args.pages, doc.pageCount + 1)
+    pages = get_list(args.pages, doc.page_count + 1)
     outdoc = fitz.open()
     for pno in pages:
         n = pno - 1
-        outdoc.insertPDF(doc, from_page=n, to_page=n)
+        outdoc.insert_pdf(doc, from_page=n, to_page=n)
     outdoc.save(
         args.output,
         garbage=args.garbage,
@@ -256,8 +267,7 @@ def clean(args):
 
 
 def doc_join(args):
-    """Join pages from several PDF documents.
-    """
+    """Join pages from several PDF documents."""
     doc_list = args.input  # a list of input PDFs
     doc = fitz.open()  # output PDF
     for src_item in doc_list:  # process one input PDF
@@ -266,11 +276,11 @@ def doc_join(args):
         src = open_file(src_list[0], password, pdf=True)
         pages = ",".join(src_list[2:])  # get 'pages' specifications
         if pages:  # if anything there, retrieve a list of desired pages
-            page_list = get_list(",".join(src_list[2:]), src.pageCount + 1)
+            page_list = get_list(",".join(src_list[2:]), src.page_count + 1)
         else:  # take all pages
-            page_list = range(1, src.pageCount + 1)
+            page_list = range(1, src.page_count + 1)
         for i in page_list:
-            doc.insertPDF(src, from_page=i - 1, to_page=i - 1)  # copy each source page
+            doc.insert_pdf(src, from_page=i - 1, to_page=i - 1)  # copy each source page
         src.close()
 
     doc.save(args.output, garbage=4, deflate=True)
@@ -278,8 +288,7 @@ def doc_join(args):
 
 
 def embedded_copy(args):
-    """Copy embedded files between PDFs.
-    """
+    """Copy embedded files between PDFs."""
     doc = open_file(args.input, args.password, pdf=True)
     if not doc.can_save_incrementally() and (
         not args.output or args.output == args.input
@@ -287,7 +296,7 @@ def embedded_copy(args):
         sys.exit("cannot save PDF incrementally")
     src = open_file(args.source, args.pwdsource)
     names = set(args.name) if args.name else set()
-    src_names = set(src.embeddedFileNames())
+    src_names = set(src.embfile_names())
     if names:
         if not names <= src_names:
             sys.exit("not all names are contained in source")
@@ -295,16 +304,14 @@ def embedded_copy(args):
         names = src_names
     if not names:
         sys.exit("nothing to copy")
-    intersect = names & set(
-        doc.embeddedFileNames()
-    )  # any equal name already in target?
+    intersect = names & set(doc.embfile_names())  # any equal name already in target?
     if intersect:
         sys.exit("following names already exist in receiving PDF: %s" % str(intersect))
 
     for item in names:
-        info = src.embeddedFileInfo(item)
-        buff = src.embeddedFileGet(item)
-        doc.embeddedFileAdd(
+        info = src.embfile_info(item)
+        buff = src.embfile_get(item)
+        doc.embfile_add(
             item,
             buff,
             filename=info["filename"],
@@ -321,8 +328,7 @@ def embedded_copy(args):
 
 
 def embedded_del(args):
-    """Delete an embedded file entry.
-    """
+    """Delete an embedded file entry."""
     doc = open_file(args.input, args.password, pdf=True)
     if not doc.can_save_incrementally() and (
         not args.output or args.output == args.input
@@ -330,23 +336,22 @@ def embedded_del(args):
         sys.exit("cannot save PDF incrementally")
 
     try:
-        doc.embeddedFileDel(args.name)
+        doc.embfile_del(args.name)
     except ValueError:
         sys.exit("no such embedded file '%s'" % args.name)
     if not args.output or args.output == args.input:
-        doc.saveIncr()
+        doc.save_incr()
     else:
         doc.save(args.output, garbage=1)
     doc.close()
 
 
 def embedded_get(args):
-    """Retrieve contents of an embedded file.
-    """
+    """Retrieve contents of an embedded file."""
     doc = open_file(args.input, args.password, pdf=True)
     try:
-        stream = doc.embeddedFileGet(args.name)
-        d = doc.embeddedFileInfo(args.name)
+        stream = doc.embfile_get(args.name)
+        d = doc.embfile_info(args.name)
     except ValueError:
         sys.exit("no such embedded file '%s'" % args.name)
     filename = args.output if args.output else d["filename"]
@@ -358,8 +363,7 @@ def embedded_get(args):
 
 
 def embedded_add(args):
-    """Insert a new embedded file.
-    """
+    """Insert a new embedded file."""
     doc = open_file(args.input, args.password, pdf=True)
     if not doc.can_save_incrementally() and (
         args.output is None or args.output == args.input
@@ -367,7 +371,7 @@ def embedded_add(args):
         sys.exit("cannot save PDF incrementally")
 
     try:
-        doc.embeddedFileDel(args.name)
+        doc.embfile_del(args.name)
         sys.exit("entry '%s' already exists" % args.name)
     except:
         pass
@@ -381,7 +385,7 @@ def embedded_add(args):
         desc = filename
     else:
         desc = args.desc
-    doc.embeddedFileAdd(
+    doc.embfile_add(
         args.name, stream, filename=filename, ufilename=ufilename, desc=desc
     )
     if not args.output or args.output == args.input:
@@ -392,8 +396,7 @@ def embedded_add(args):
 
 
 def embedded_upd(args):
-    """Update contents or metadata of an embedded file.
-    """
+    """Update contents or metadata of an embedded file."""
     doc = open_file(args.input, args.password, pdf=True)
     if not doc.can_save_incrementally() and (
         args.output is None or args.output == args.input
@@ -401,7 +404,7 @@ def embedded_upd(args):
         sys.exit("cannot save PDF incrementally")
 
     try:
-        doc.embeddedFileInfo(args.name)
+        doc.embfile_info(args.name)
     except:
         sys.exit("no such embedded file '%s'" % args.name)
 
@@ -431,7 +434,7 @@ def embedded_upd(args):
     else:
         desc = None
 
-    doc.embeddedFileUpd(
+    doc.embfile_upd(
         args.name, stream, filename=filename, ufilename=ufilename, desc=desc
     )
     if args.output is None or args.output == args.input:
@@ -442,10 +445,9 @@ def embedded_upd(args):
 
 
 def embedded_list(args):
-    """List embedded files.
-    """
+    """List embedded files."""
     doc = open_file(args.input, args.password, pdf=True)
-    names = doc.embeddedFileNames()
+    names = doc.embfile_names()
     if args.name is not None:
         if args.name not in names:
             sys.exit("no such embedded file '%s'" % args.name)
@@ -456,7 +458,7 @@ def embedded_list(args):
                 % (len(names), "s" if len(names) > 1 else "")
             )
             print()
-            print_dict(doc.embeddedFileInfo(args.name))
+            print_dict(doc.embfile_info(args.name))
             print()
             return
     if not names:
@@ -472,23 +474,22 @@ def embedded_list(args):
         if not args.detail:
             print(name)
             continue
-        _ = doc.embeddedFileInfo(name)
-        print_dict(doc.embeddedFileInfo(name))
+        _ = doc.embfile_info(name)
+        print_dict(doc.embfile_info(name))
         print()
     doc.close()
 
 
 def extract_objects(args):
-    """Extract images and / or fonts from a PDF.
-    """
+    """Extract images and / or fonts from a PDF."""
     if not args.fonts and not args.images:
         sys.exit("neither fonts nor images requested")
     doc = open_file(args.input, args.password, pdf=True)
 
     if args.pages:
-        pages = get_list(args.pages, doc.pageCount + 1)
+        pages = get_list(args.pages, doc.page_count + 1)
     else:
-        pages = range(1, doc.pageCount + 1)
+        pages = range(1, doc.page_count + 1)
 
     if not args.output:
         out_dir = os.path.abspath(os.curdir)
@@ -502,12 +503,12 @@ def extract_objects(args):
 
     for pno in pages:
         if args.fonts:
-            itemlist = doc.getPageFontList(pno - 1)
+            itemlist = doc.get_page_fonts(pno - 1)
             for item in itemlist:
                 xref = item[0]
                 if xref not in font_xrefs:
                     font_xrefs.add(xref)
-                    fontname, ext, _, buffer = doc.extractFont(xref)
+                    fontname, ext, _, buffer = doc.extract_font(xref)
                     if ext == "n/a" or not buffer:
                         continue
                     outname = os.path.join(
@@ -518,7 +519,7 @@ def extract_objects(args):
                     outfile.close()
                     buffer = None
         if args.images:
-            itemlist = doc.getPageImageList(pno - 1)
+            itemlist = doc.get_page_images(pno - 1)
             for item in itemlist:
                 xref = item[0]
                 if xref not in image_xrefs:
@@ -538,7 +539,7 @@ def extract_objects(args):
                             if pix.colorspace.n < 4
                             else fitz.Pixmap(fitz.csRGB, pix)
                         )
-                        pix2.writeImage(outname)
+                        pix2.save(outname)
 
     if args.fonts:
         print("saved %i fonts to '%s'" % (len(font_xrefs), out_dir))
@@ -547,13 +548,297 @@ def extract_objects(args):
     doc.close()
 
 
-def main():
-    """Define command configurations.
-    """
-    import argparse
+def page_simple(page, textout, GRID, fontsize, noformfeed, skip_empty, flags):
+    eop = b"\n" if noformfeed else bytes([12])
+    text = page.get_text("text", flags=flags)
+    if not text:
+        if not skip_empty:
+            textout.write(eop)  # write formfeed
+        return
+    textout.write(text.encode("utf8", errors="surrogatepass"))
+    textout.write(eop)
+    return
 
+
+def page_blocksort(page, textout, GRID, fontsize, noformfeed, skip_empty, flags):
+    eop = b"\n" if noformfeed else bytes([12])
+    blocks = page.get_text("blocks", flags=flags)
+    if blocks == []:
+        if not skip_empty:
+            textout.write(eop)  # write formfeed
+        return
+    blocks.sort(key=lambda b: (b[3], b[0]))
+    for b in blocks:
+        textout.write(b[4].encode("utf8", errors="surrogatepass"))
+    textout.write(eop)
+    return
+
+
+def page_layout(page, textout, GRID, fontsize, noformfeed, skip_empty, flags):
+    eop = b"\n" if noformfeed else bytes([12])
+
+    # --------------------------------------------------------------------
+    def find_line_index(values: List[int], value: int) -> int:
+        """Find the right row coordinate.
+
+        Args:
+            values: (list) y-coordinates of rows.
+            value: (int) lookup for this value (y-origin of char).
+        Returns:
+            y-ccordinate of appropriate line for value.
+        """
+        i = bisect.bisect_right(values, value)
+        if i:
+            return values[i - 1]
+        raise RuntimeError("Line for %g not found in %s" % (value, values))
+
+    # --------------------------------------------------------------------
+    def curate_rows(rows: Set[int], GRID) -> List:
+        rows = list(rows)
+        rows.sort()  # sort ascending
+        nrows = [rows[0]]
+        for h in rows[1:]:
+            if h >= nrows[-1] + GRID:  # only keep significant differences
+                nrows.append(h)
+        return nrows  # curated list of line bottom coordinates
+
+    def process_blocks(blocks: List[Dict], page: fitz.Page):
+        rows = set()
+        page_width = page.rect.width
+        page_height = page.rect.height
+        rowheight = page_height
+        left = page_width
+        right = 0
+        chars = []
+        for block in blocks:
+            for line in block["lines"]:
+                if line["dir"] != (1, 0):  # ignore non-horizontal text
+                    continue
+                x0, y0, x1, y1 = line["bbox"]
+                if y1 < 0 or y0 > page.rect.height:  # ignore if outside CropBox
+                    continue
+                # upd row height
+                height = y1 - y0
+
+                if rowheight > height:
+                    rowheight = height
+                for span in line["spans"]:
+                    if span["size"] <= fontsize:
+                        continue
+                    for c in span["chars"]:
+                        x0, _, x1, _ = c["bbox"]
+                        cwidth = x1 - x0
+                        ox, oy = c["origin"]
+                        oy = int(round(oy))
+                        rows.add(oy)
+                        ch = c["c"]
+                        if left > ox and ch != " ":
+                            left = ox  # update left coordinate
+                        if right < x1:
+                            right = x1  # update right coordinate
+                        # handle ligatures:
+                        if cwidth == 0 and chars != []:  # potential ligature
+                            old_ch, old_ox, old_oy, old_cwidth = chars[-1]
+                            if old_oy == oy:  # ligature
+                                if old_ch != chr(0xFB00):  # previous "ff" char lig?
+                                    lig = joinligature(old_ch + ch)  # no
+                                # convert to one of the 3-char ligatures:
+                                elif ch == "i":
+                                    lig = chr(0xFB03)  # "ffi"
+                                elif ch == "l":
+                                    lig = chr(0xFB04)  # "ffl"
+                                else:  # something wrong, leave old char in place
+                                    lig = old_ch
+                                chars[-1] = (lig, old_ox, old_oy, old_cwidth)
+                                continue
+                        chars.append((ch, ox, oy, cwidth))  # all chars on page
+        return chars, rows, left, right, rowheight
+
+    def joinligature(lig: str) -> str:
+        """Return ligature character for a given pair / triple of characters.
+
+        Args:
+            lig: (str) 2/3 characters, e.g. "ff"
+        Returns:
+            Ligature, e.g. "ff" -> chr(0xFB00)
+        """
+
+        if lig == "ff":
+            return chr(0xFB00)
+        elif lig == "fi":
+            return chr(0xFB01)
+        elif lig == "fl":
+            return chr(0xFB02)
+        elif lig == "ffi":
+            return chr(0xFB03)
+        elif lig == "ffl":
+            return chr(0xFB04)
+        elif lig == "ft":
+            return chr(0xFB05)
+        elif lig == "st":
+            return chr(0xFB06)
+        return lig
+
+    # --------------------------------------------------------------------
+    def make_textline(left, slot, minslot, lchars):
+        """Produce the text of one output line.
+
+        Args:
+            left: (float) left most coordinate used on page
+            slot: (float) avg width of one character in any font in use.
+            minslot: (float) min width for the characters in this line.
+            chars: (list[tuple]) characters of this line.
+        Returns:
+            text: (str) text string for this line
+        """
+        text = ""  # we output this
+        old_char = ""
+        old_x1 = 0  # end coordinate of last char
+        old_ox = 0  # x-origin of last char
+        if minslot <= fitz.EPSILON:
+            raise RuntimeError("program error: minslot too small = %g" % minslot)
+
+        for c in lchars:  # loop over characters
+            char, ox, _, cwidth = c
+            ox = ox - left  # its (relative) start coordinate
+            x1 = ox + cwidth  # ending coordinate
+
+            # eliminate overprint effect
+            if old_char == char and ox - old_ox <= cwidth * 0.2:
+                continue
+
+            # omit spaces overlapping previous char
+            if char == " " and (old_x1 - ox) / cwidth > 0.8:
+                continue
+
+            old_char = char
+            # close enough to previous?
+            if ox < old_x1 + minslot:  # assume char adjacent to previous
+                text += char  # append to output
+                old_x1 = x1  # new end coord
+                old_ox = ox  # new origin.x
+                continue
+
+            # else next char starts after some gap:
+            # fill in right number of spaces, so char is positioned
+            # in the right slot of the line
+            if char == " ":  # rest relevant for non-space only
+                continue
+            delta = int(ox / slot) - len(text)
+            if ox > old_x1 and delta > 1:
+                text += " " * delta
+            # now append char
+            text += char
+            old_x1 = x1  # new end coordinate
+            old_ox = ox  # new origin
+        return text.rstrip()
+
+    # extract page text by single characters ("rawdict")
+    blocks = page.get_text("rawdict", flags=flags)["blocks"]
+    chars, rows, left, right, rowheight = process_blocks(blocks, page)
+
+    if chars == []:
+        if not skip_empty:
+            textout.write(eop)  # write formfeed
+        return
+    # compute list of line coordinates - ignoring small (GRID) differences
+    rows = curate_rows(rows, GRID)
+
+    # sort all chars by x-coordinates, so every line will receive char info,
+    # sorted from left to right.
+    chars.sort(key=lambda c: c[1])
+
+    # populate the lines with their char info
+    lines = {}  # key: y1-ccordinate, value: char list
+    for c in chars:
+        _, _, oy, _ = c
+        y = find_line_index(rows, oy)  # y-coord of the right line
+        lchars = lines.get(y, [])  # read line chars so far
+        lchars.append(c)  # append this char
+        lines[y] = lchars  # write back to line
+
+    # ensure line coordinates are ascending
+    keys = list(lines.keys())
+    keys.sort()
+
+    # -------------------------------------------------------------------------
+    # Compute "char resolution" for the page: the char width corresponding to
+    # 1 text char position on output - call it 'slot'.
+    # For each line, compute median of its char widths. The minimum across all
+    # lines is 'slot'.
+    # The minimum char width of each line is used to determine if spaces must
+    # be inserted in between two characters.
+    # -------------------------------------------------------------------------
+    slot = right - left
+    minslots = {}
+    for k in keys:
+        lchars = lines[k]
+        ccount = len(lchars)
+        if ccount < 2:
+            minslots[k] = 1
+            continue
+        widths = [c[3] for c in lchars]
+        widths.sort()
+        this_slot = statistics.median(widths)  # take median value
+        if this_slot < slot:
+            slot = this_slot
+        minslots[k] = widths[0]
+
+    # compute line advance in text output
+    rowheight = rowheight * (rows[-1] - rows[0]) / (rowheight * len(rows)) * 1.2
+    rowpos = rows[0]  # first line positioned here
+    textout.write(b"\n")
+    for k in keys:  # walk through the lines
+        while rowpos < k:  # honor distance between lines
+            textout.write(b"\n")
+            rowpos += rowheight
+        text = make_textline(left, slot, minslots[k], lines[k])
+        textout.write((text + "\n").encode("utf8", errors="surrogatepass"))
+        rowpos = k + rowheight
+
+    textout.write(eop)  # write formfeed
+
+
+def gettext(args):
+    doc = open_file(args.input, args.password, pdf=False)
+    pagel = get_list(args.pages, doc.page_count + 1)
+    output = args.output
+    if output == None:
+        filename, _ = os.path.splitext(doc.name)
+        output = filename + ".txt"
+    textout = open(output, "wb")
+    flags = TEXT_PRESERVE_LIGATURES | TEXT_PRESERVE_WHITESPACE
+    if args.convert_white:
+        flags ^= TEXT_PRESERVE_WHITESPACE
+    if args.noligatures:
+        flags ^= TEXT_PRESERVE_LIGATURES
+    if args.extra_spaces:
+        flags ^= TEXT_INHIBIT_SPACES
+    func = {
+        "simple": page_simple,
+        "blocks": page_blocksort,
+        "layout": page_layout,
+    }
+    for pno in pagel:
+        page = doc[pno - 1]
+        func[args.mode](
+            page,
+            textout,
+            args.grid,
+            args.fontsize,
+            args.noformfeed,
+            args.skip_empty,
+            flags=flags,
+        )
+
+    textout.close()
+
+
+def main():
+    """Define command configurations."""
     parser = argparse.ArgumentParser(
-        description=mycenter("Basic PyMuPDF Functions"), prog="fitz"
+        prog="fitz",
+        description=mycenter("Basic PyMuPDF Functions"),
     )
     subps = parser.add_subparsers(
         title="Subcommands", help="Enter 'command -h' for subcommand specific help"
@@ -580,7 +865,7 @@ def main():
     # 'clean' command
     # -------------------------------------------------------------------------
     ps_clean = subps.add_parser(
-        "clean", description=mycenter("optimize PDF or create sub-PDF if pages given")
+        "clean", description=mycenter("optimize PDF, or create sub-PDF if pages given")
     )
     ps_clean.add_argument("input", type=str, help="PDF filename")
     ps_clean.add_argument("output", type=str, help="output PDF filename")
@@ -768,6 +1053,75 @@ def main():
     )
     ps_embed_copy.set_defaults(func=embedded_copy)
 
+    # -------------------------------------------------------------------------
+    # 'textlayout' command
+    # -------------------------------------------------------------------------
+    ps_gettext = subps.add_parser(
+        "gettext", description=mycenter("extract text in various formatting modes")
+    )
+    ps_gettext.add_argument("input", type=str, help="input document filename")
+    ps_gettext.add_argument("-password", help="password for input document")
+    ps_gettext.add_argument(
+        "-mode",
+        type=str,
+        help="mode: simple, block sort, or layout (default)",
+        choices=("simple", "blocks", "layout"),
+        default="layout",
+    )
+    ps_gettext.add_argument(
+        "-pages",
+        type=str,
+        help="select pages, format: 1,5-7,50-N",
+        default="1-N",
+    )
+    ps_gettext.add_argument(
+        "-noligatures",
+        action="store_true",
+        help="expand ligature characters (default False)",
+        default=False,
+    )
+    ps_gettext.add_argument(
+        "-convert-white",
+        action="store_true",
+        help="convert whitespace characters to white (default False)",
+        default=False,
+    )
+    ps_gettext.add_argument(
+        "-extra-spaces",
+        action="store_true",
+        help="fill gaps with spaces (default False)",
+        default=False,
+    )
+    ps_gettext.add_argument(
+        "-noformfeed",
+        action="store_true",
+        help="write linefeeds, no formfeeds (default False)",
+        default=False,
+    )
+    ps_gettext.add_argument(
+        "-skip-empty",
+        action="store_true",
+        help="suppress pages with no text (default False)",
+        default=False,
+    )
+    ps_gettext.add_argument(
+        "-output",
+        help="store text in this file (default inputfilename.txt)",
+    )
+    ps_gettext.add_argument(
+        "-grid",
+        type=float,
+        help="merge lines if closer than this (default 2)",
+        default=2,
+    )
+    ps_gettext.add_argument(
+        "-fontsize",
+        type=float,
+        help="only include text with a larger fontsize (default 3)",
+        default=3,
+    )
+    ps_gettext.set_defaults(func=gettext)
+
     # -------------------------------------------------------------------------
     # start program
     # -------------------------------------------------------------------------
